// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    expr::Target,
    intern::{Intern, Interned},
};
use serde::{Deserialize, Serialize};
use std::{
    fmt,
    hash::{Hash, Hasher},
    ops::Deref,
};

#[derive(Clone)]
struct CustomFirrtlAnnotationFieldsImpl {
    value: serde_json::Map<String, serde_json::Value>,
    serialized: Interned<str>,
}

impl Hash for CustomFirrtlAnnotationFieldsImpl {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.serialized.hash(state);
    }
}

impl Eq for CustomFirrtlAnnotationFieldsImpl {}

impl PartialEq for CustomFirrtlAnnotationFieldsImpl {
    fn eq(&self, other: &Self) -> bool {
        self.serialized == other.serialized
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct CustomFirrtlAnnotationFields(Interned<CustomFirrtlAnnotationFieldsImpl>);

impl fmt::Debug for CustomFirrtlAnnotationFields {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.value.fmt(f)
    }
}

impl<'de> Deserialize<'de> for CustomFirrtlAnnotationFields {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        serde_json::Map::<String, serde_json::Value>::deserialize(deserializer).map(Self::from)
    }
}

impl Serialize for CustomFirrtlAnnotationFields {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        self.0.value.serialize(serializer)
    }
}

impl Deref for CustomFirrtlAnnotationFields {
    type Target = serde_json::Map<String, serde_json::Value>;

    fn deref(&self) -> &Self::Target {
        &self.0.value
    }
}

impl From<serde_json::Map<String, serde_json::Value>> for CustomFirrtlAnnotationFields {
    fn from(value: serde_json::Map<String, serde_json::Value>) -> Self {
        let serialized =
            serde_json::to_string(&value).expect("serialization of JSON should succeed");
        Self(Intern::intern_sized(CustomFirrtlAnnotationFieldsImpl {
            value,
            serialized: Intern::intern_owned(serialized),
        }))
    }
}

#[derive(Debug, Clone)]
pub struct NotAJsonObject(pub serde_json::Value);

impl fmt::Display for NotAJsonObject {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str("not a JSON object")
    }
}

impl std::error::Error for NotAJsonObject {}

impl TryFrom<serde_json::Value> for CustomFirrtlAnnotationFields {
    type Error = NotAJsonObject;

    fn try_from(value: serde_json::Value) -> Result<Self, Self::Error> {
        match value {
            serde_json::Value::Object(value) => Ok(value.into()),
            _ => Err(NotAJsonObject(value)),
        }
    }
}

impl From<CustomFirrtlAnnotationFields> for serde_json::Map<String, serde_json::Value> {
    fn from(value: CustomFirrtlAnnotationFields) -> Self {
        Self::clone(&value)
    }
}

impl From<CustomFirrtlAnnotationFields> for serde_json::Value {
    fn from(value: CustomFirrtlAnnotationFields) -> Self {
        serde_json::Value::Object(value.into())
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Debug, Hash, Serialize, Deserialize)]
pub struct CustomFirrtlAnnotation {
    pub class: Interned<str>,
    #[serde(flatten)]
    pub additional_fields: CustomFirrtlAnnotationFields,
}

#[derive(Clone, PartialEq, Eq, Hash, Debug)]
#[non_exhaustive]
pub enum Annotation {
    DontTouch,
    CustomFirrtl(CustomFirrtlAnnotation),
}

#[derive(Clone, PartialEq, Eq, Hash, Debug)]
pub struct TargetedAnnotation {
    target: Interned<Target>,
    annotation: Annotation,
}

impl TargetedAnnotation {
    #[track_caller]
    pub fn new(target: Interned<Target>, annotation: Annotation) -> Self {
        Self::assert_valid_target(target);
        Self { target, annotation }
    }
    #[track_caller]
    pub fn assert_valid_target(target: Interned<Target>) {
        assert!(target.is_static(), "can't annotate non-static targets");
    }
    pub fn target(&self) -> Interned<Target> {
        self.target
    }
    pub fn annotation(&self) -> &Annotation {
        &self.annotation
    }
}

pub trait IntoAnnotations {
    type IntoAnnotations: IntoIterator<Item = Annotation>;

    fn into_annotations(self) -> Self::IntoAnnotations;
}

impl IntoAnnotations for Annotation {
    type IntoAnnotations = [Annotation; 1];

    fn into_annotations(self) -> Self::IntoAnnotations {
        [self]
    }
}

impl IntoAnnotations for Box<Annotation> {
    type IntoAnnotations = [Annotation; 1];

    fn into_annotations(self) -> Self::IntoAnnotations {
        [*self]
    }
}

impl IntoAnnotations for &'_ Annotation {
    type IntoAnnotations = [Annotation; 1];

    fn into_annotations(self) -> Self::IntoAnnotations {
        [self.clone()]
    }
}

impl IntoAnnotations for &'_ mut Annotation {
    type IntoAnnotations = [Annotation; 1];

    fn into_annotations(self) -> Self::IntoAnnotations {
        [self.clone()]
    }
}

impl<T: IntoIterator<Item = Annotation>> IntoAnnotations for T {
    type IntoAnnotations = Self;

    fn into_annotations(self) -> Self::IntoAnnotations {
        self
    }
}
